#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <curl/curl.h>
#include <jansson.h>
#include "musicbrainz-lib.h"

struct MBState * mbstate_alloc(void)
{
  struct MBState * mbs ;

  mbs = malloc(sizeof(struct MBState)) ;
  mbs->curl_handle = 0;
  mbs->logged_in = 0;
  return(mbs);
}

void mb_global_initialize()
{
  int cgi_ret;
  /*  curl_global_init(CURL_GLOBAL_WIN32 | CURL_GLOBAL_SSL ) */
  cgi_ret = curl_global_init(CURL_GLOBAL_SSL) ;
  if(cgi_ret != 0)
  {
    printf("curl_global_init() failed\n");
    abort();
  }
}

void mbstate_initialize(struct MBState * mbs)
{
  CURL * ch;

  ch = curl_easy_init() ;
  mbs->curl_handle = ch;
}

struct MBState * mb_initialize(void)
{
  CURL * ch;
  struct MBState * mbs ;

  mbs = mbstate_alloc();
  mbstate_initialize(mbs);

  return(mbs);
}

size_t login_header_callback(char *buffer,
		       size_t size,
		       size_t nitems,
		       void *userdata)
{
  int i;
  return(size * nitems);
}

size_t add_artist_header_callback(char *buffer,
		       size_t size,
		       size_t nitems,
		       void *userdata)
{
  int i;
  return(size * nitems);
}

size_t login_write_callback(char *buffer,
			    size_t size,
			    size_t nitems,
			    void *userdata)
{
  return(size * nitems);
}

size_t add_artist_write_callback(char *buffer,
			    size_t size,
			    size_t nitems,
			    void *userdata)
{
  return(size * nitems);
}

size_t do_nothing_write_callback(char *buffer,
			    size_t size,
			    size_t nitems,
			    void *userdata)
{
  return(size * nitems);
}

size_t build_string_write_callback(char *buffer,
			    size_t size,
			    size_t nitems,
			    void *userdata)
{
  char * sofar;
  char * output;
  size_t oldsize, newsize;
  char ** ud;

  ud = (char **) userdata;

  if (userdata == NULL)
    return(size * nitems);

  /*
  sofar = (char *) *userdata;
  sofar = *ud;
  */
  sofar = * (char **) userdata;

  oldsize = (sofar == NULL ? 0 : strlen(sofar));

  newsize = oldsize + size * nitems + 1;
  sofar = realloc(sofar, newsize);
  sofar[oldsize] = 0;
  strncat(sofar, buffer, size * nitems);
  sofar[oldsize + size * nitems] = 0;

  /*
  * (char **) userdata = sofar;
  */
  *ud = sofar;
  return(size * nitems);
}

int nullify_curl_options(CURL * curl, int count, CURLoption options[])
{
  int i;
  for(i = 0; i < count ; i++)
  {
    switch(options[i]) {
    case CURLOPT_WRITEDATA :
      curl_easy_setopt(curl, options[i], stdout);
      break;
    default :
      curl_easy_setopt(curl, options[i], NULL);
    }
  }
}

int mb_login(struct MBState * mbs, char* username, char* password)
{
  CURLcode ret;

  char * postfields;
  int postfields_length;
  int numchars;
  char * header = NULL;
  long int status_code = 0;
  char * redirect_url;

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_URL,
			 "https://musicbrainz.org/login");
  postfields_length = strlen(username) + strlen(password) + 20 ;
  postfields = malloc(postfields_length) ;

  assert(sprintf(postfields, "username=%s&password=%s", username, password) <
	 postfields_length) ;
 
  /* permanent options */
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_COOKIEFILE, "");

  /* options for just this transfer */
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_POSTFIELDS, postfields) ;
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_HEADERDATA, &header) ;
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_HEADERFUNCTION,
			 &login_header_callback);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEFUNCTION,
			 &login_write_callback);

  ret = curl_easy_perform(mbs->curl_handle);

  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_RESPONSE_CODE,
			  &status_code);
  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_REDIRECT_URL,
			  &redirect_url);

  /*
  printf("  status_code is %ld\n", status_code);
  printf("  redirect_url is %s\n", redirect_url);
  */

  CURLoption opts[] = {CURLOPT_HEADERFUNCTION, CURLOPT_POSTFIELDS,
			   CURLOPT_HEADERDATA, CURLOPT_WRITEFUNCTION};
  nullify_curl_options(mbs->curl_handle, 4, opts);

  free((void *) postfields);
  if(status_code == 302)
  {
    mbs->logged_in = 1;
    return(0);
  }
  else
  {
    mbs->logged_in = 0;
    return(1);
  }
}

char * build_form_chunk(CURL * ch, char * field_name, char * field_val)
{
  char * field_val_encoded;
  char * out;
  int postfields_length;
  CURLcode ret;

  field_val_encoded = curl_easy_escape(ch, field_val, 0);

  char * postformat = "%s=%s";
  postfields_length = strlen(postformat) + strlen(field_name) +
    strlen(field_val_encoded);
  out = malloc(postfields_length);
  assert(sprintf(out, postformat, field_name, field_val_encoded) <
	 postfields_length) ;

  curl_free(field_val_encoded);

  return(out);
}

char * itoa (int num) {
  int len = 0;
  int x;
  char * out;

  for (x = num; x != 0; x = x / 10)
    len = len + 1;

  if(num < 0)
    len = len + 1;

  out = malloc(len + 1);
  sprintf(out, "%d", num);
  return(out);
  
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
char * add_chunk_to_postfields(char * postfields, char * chunk)
{
  char * postfields2;
  int postfields_length = strlen(postfields);

  postfields_length += (strlen(chunk) + 2) ;

  postfields2 = malloc(postfields_length) ;
  sprintf(postfields2, "%s&%s", postfields, chunk);

  free(postfields);

  return(postfields2);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(*postfields) or a realloc
*/
void add_chunk_to_postfields3(char ** postfields, char * chunk)
{
  char * postfields2;
  int postfields_length = strlen(*postfields);

  postfields_length += (strlen(chunk) + 2) ;

  postfields2 = malloc(postfields_length) ;
  sprintf(postfields2, "%s&%s", *postfields, chunk);

  free(*postfields);

  *postfields = postfields2;
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
char * add_chunk_to_postfields2(CURL * curl, char * postfields,
				char * field_name, char * field_val)
{
  char * newchunk;

  newchunk = build_form_chunk(curl, field_name, field_val);
  postfields = add_chunk_to_postfields(postfields, newchunk);
  /*
  add_chunk_to_postfields3(&postfields, newchunk);
  */
  free(newchunk);

  return(postfields);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(*postfields) or a realloc
*/
void add_chunk_to_postfields4(CURL * curl, char ** postfields,
				char * field_name, char * field_val)
{
  char * newchunk;

  newchunk = build_form_chunk(curl, field_name, field_val);
  *postfields = add_chunk_to_postfields(*postfields, newchunk);
  /*
  add_chunk_to_postfields3(postfields, newchunk);
  */
  free(newchunk);

}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(*postfields) or a realloc
*/
void add_chunk_to_postfields5(CURL * curl, char ** postfields,
				char * field_name, int field_val)
{
  char * newchunk;
  char * field_val_string;

  field_val_string = itoa(field_val);

  newchunk = build_form_chunk(curl, field_name, field_val_string);
  *postfields = add_chunk_to_postfields(*postfields, newchunk);
  /*
  add_chunk_to_postfields3(postfields, newchunk);
  */

  free(field_val_string);
  free(newchunk);

}

int add_numbered_chunk_to_postfields(CURL * curl, char ** postfields,
				     char * field, int num, char * value)
{
  char * field2;

  field2 = malloc(strlen(field) + num);

  sprintf(field2, field, num);

  add_chunk_to_postfields4(curl, postfields, field2, value);

  free(field2);
  
  return(0);
}

int add_numbered_chunk_to_postfields2(CURL * curl, char ** postfields,
				      char * buf,
				      char * field, int num, char * value)
{
  char * field2 = buf;

  sprintf(buf, field, num);

  add_chunk_to_postfields4(curl, postfields, buf, value);
  
  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_type_to_artist(CURL *curl, char ** postfields,
		       struct MBArtistInfo artist)
{
  if (artist.type != 0)
    add_chunk_to_postfields5(curl, postfields,
			     "edit-artist.type_id", artist.type);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_gender_to_artist(CURL *curl, char ** postfields,
			 struct MBArtistInfo artist)
{
  if (artist.gender != 0)
    add_chunk_to_postfields5(curl, postfields,
			     "edit-artist.gender_id", artist.gender);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_disambig_to_artist(CURL *curl, char ** postfields,
			  struct MBArtistInfo artist)
{
  if (artist.disambiguation == NULL)
    return(0);

  add_chunk_to_postfields4(curl, postfields,
			   "edit-artist.comment", artist.disambiguation);

  if(artist.not_dupe == 1)
    add_chunk_to_postfields3(postfields, "edit-artist.not_dupe=1");

  return(0);
}

/*
reads area_id field of the "artist" parameter, which is an int, and adds the
appropriate field to postfields.

assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_area_to_artist(CURL *curl, char ** postfields,
			  struct MBArtistInfo artist)
{
  if (artist.area_id != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-artist.area_id", artist.area_id);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_urls_to_artist(CURL *curl, char ** postfields,
			  struct MBArtistInfo artist)
{
  char * lt_field_name;
  char * text_field_name;

  int i;

  char * ltfnf = "edit-artist.url.%d.link_type_id";
  char * tfnf = "edit-artist.url.%d.text";

  lt_field_name = malloc(strlen(ltfnf) + artist.url_count);
  text_field_name = malloc(strlen(tfnf) + artist.url_count);

  for (i = 0; i < artist.url_count && artist.urls; i++)
  {
    sprintf(lt_field_name, ltfnf, i);
    sprintf(text_field_name, tfnf, i);

    add_chunk_to_postfields5(curl, postfields,
			     lt_field_name, artist.urls[i].type);

    add_chunk_to_postfields4(curl, postfields,
			     text_field_name, artist.urls[i].address);
  }

  free(lt_field_name);
  free(text_field_name);
  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_relationships_to_artist(CURL *curl, char ** postfields,
				   struct MBArtistInfo artist)
{
  char * bwfield;
  char * typefield;
  char * targetfield;

  int i;

  char * bwfieldformat = "edit-artist.rel.%d.backward";
  bwfield = malloc(strlen(bwfieldformat) + artist.relationship_count);
  char * typefieldformat = "edit-artist.rel.%d.link_type_id";
  typefield = malloc(strlen(typefieldformat) + artist.relationship_count);
  char * targetfieldformat = "edit-artist.rel.%d.target";
  targetfield = malloc(strlen(targetfieldformat) + artist.relationship_count);

  for (i = 0; i < artist.relationship_count && artist.relationships; i++)
  {
    if (artist.relationships[i].backward)
    {
      sprintf(bwfield, bwfieldformat, i);

      add_chunk_to_postfields4(curl, postfields, bwfield, "1");
    }

    sprintf(typefield, typefieldformat, i);

    add_chunk_to_postfields5(curl, postfields,
			     typefield, artist.relationships[i].type);

    sprintf(targetfield, targetfieldformat, i);

    add_chunk_to_postfields4(curl, postfields, targetfield,
			     artist.relationships[i].musicbrainzID);
  }
  free(bwfield);
  free(typefield);
  free(targetfield);

  return(0);
}

/*
This will probably puke/crash if artist_name or sort_name field is NULL
 */
char * build_artist_name_post_string(CURL * ch, struct MBArtistInfo artist)
{
  char * postfields;
  char * sortname_chunk;

  postfields = build_form_chunk(ch, "edit-artist.name", artist.artist_name);

  sortname_chunk = build_form_chunk(ch, "edit-artist.sort_name",
				    artist.sort_name);
  postfields = add_chunk_to_postfields(postfields, sortname_chunk);

  free(sortname_chunk);

  return(postfields);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_disambiguation_to_place(CURL *curl, char ** postfields,
				   struct MBPlace place)
{
  if (place.disambiguation != NULL)
    add_chunk_to_postfields4(curl, postfields,
			     "edit-place.comment", place.disambiguation);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_dupe_to_place(CURL *curl, char ** postfields, struct MBPlace place)
{
  if (place.not_dupe == 1)
    add_chunk_to_postfields4(curl, postfields, "edit-place.not_dupe", "1");

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_type_to_place(CURL *curl, char ** postfields, struct MBPlace place)
{
  if (place.type != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-place.type_id", place.type);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_coordinates_to_place(CURL *curl, char ** postfields,
				struct MBPlace place)
{
  char * coords;

  if (place.coordinates == NULL)
    coords = "";
  else
    coords = place.coordinates;

  add_chunk_to_postfields4(curl, postfields, "edit-place.coordinates", coords);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_address_to_place(CURL *curl, char ** postfields,
			    struct MBPlace place)
{
  if (place.address != NULL)
    add_chunk_to_postfields4(curl, postfields,
			     "edit-place.address", place.address);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_area_id_to_place(CURL *curl, char ** postfields,
			    struct MBPlace place)
{
  if (place.area_id != 0)
    add_chunk_to_postfields5(curl, postfields,
			     "edit-place.area_id", place.area_id);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_relationships_to_place(CURL * curl, char ** postfields,
				  struct MBPlace place)
{
  char * bwfield;
  char * typefield;
  char * targetfield;

  int i;

  char * bwfieldformat = "edit-place.rel.%d.backward";
  bwfield = malloc(strlen(bwfieldformat) + place.relationship_count);
  char * typefieldformat = "edit-place.rel.%d.link_type_id";
  typefield = malloc(strlen(typefieldformat) + place.relationship_count);
  char * targetfieldformat = "edit-place.rel.%d.target";
  targetfield = malloc(strlen(targetfieldformat) + place.relationship_count);

  for (i = 0; i < place.relationship_count && place.relationships; i++)
  {
    if (place.relationships[i].backward)
    {
      sprintf(bwfield, bwfieldformat, i);

      add_chunk_to_postfields4(curl, postfields, bwfield, "1");
    }

    sprintf(typefield, typefieldformat, i);
    add_chunk_to_postfields5(curl, postfields,
			     typefield, place.relationships[i].type);

    sprintf(targetfield, targetfieldformat, i);

    add_chunk_to_postfields4(curl, postfields, targetfield,
			     place.relationships[i].musicbrainzID);
  }
  free(bwfield);
  free(typefield);
  free(targetfield);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_urls_to_place(CURL *curl, char ** postfields,
			  struct MBPlace place)
{
  char * ulti_field;
  char * ut_field;

  int i;

  char * ulti_format = "edit-place.url.%d.link_type_id";
  char * ut_format = "edit-place.url.%d.text";

  ulti_field = malloc(strlen(ulti_format) + place.url_count);
  ut_field = malloc(strlen(ut_format) + place.url_count);

  for (i = 0; i < place.url_count && place.urls; i++)
  {
    sprintf(ulti_field, ulti_format, i);
    sprintf(ut_field, ut_format, i);

    add_chunk_to_postfields5(curl, postfields, ulti_field, place.urls[i].type);

    add_chunk_to_postfields4(curl, postfields,
			     ut_field, place.urls[i].address);
  }
  free(ulti_field);
  free(ut_field);

  return(0);
}

char * build_place_name_post_string(CURL * ch, struct MBPlace place)
{
  return(build_form_chunk(ch, "edit-place.name", place.name));
}

char * build_event_name_post_string(CURL * ch, struct MBEventInfo event)
{
  return(build_form_chunk(ch, "edit-event.name", event.name));
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_setlist_to_event(CURL *curl, char ** postfields,
				struct MBEventInfo event)
{
  char * setlist = (event.setlist == NULL ? "" : event.setlist);

  add_chunk_to_postfields4(curl, postfields, "edit-event.setlist", setlist);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_disambig_to_event(CURL *curl, char ** postfields,
			     struct MBEventInfo event)
{
  char * disambig;

  if (event.disambiguation == NULL)
    disambig = "";
  else
    disambig = event.disambiguation;

  add_chunk_to_postfields4(curl, postfields, "edit-event.comment", disambig);

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_type_to_event(CURL * curl, char ** postfields,
			 struct MBEventInfo event)
{
  if (event.type != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.type_id", event.type);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_begin_year_to_event(CURL * curl, char ** postfields,
			       struct MBEventInfo event)
{
  if (event.begin_year != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.begin_date.year",
			     event.begin_year);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_begin_month_to_event(CURL * curl, char ** postfields,
			     struct MBEventInfo event)
{
  if (event.begin_month != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.begin_date.month",
			     event.begin_month);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_begin_day_to_event(CURL * curl, char ** postfields,
			      struct MBEventInfo event)
{
  if (event.begin_day != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.begin_date.day",
			     event.begin_day);
  }
  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_end_year_to_event(CURL * curl, char ** postfields,
			       struct MBEventInfo event)
{
  if (event.end_year != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.end_date.year",
			     event.end_year);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_end_month_to_event(CURL * curl, char ** postfields,
			     struct MBEventInfo event)
{
  if (event.end_month != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.end_date.month",
			     event.end_month);
  }

  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_end_day_to_event(CURL * curl, char ** postfields,
			      struct MBEventInfo event)
{
  if (event.end_day != 0)
  {
    add_chunk_to_postfields5(curl, postfields,
			     "edit-event.period.end_date.day",
			     event.end_day);
  }
  return(0);
}

/*
assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_time_to_event(CURL * curl, char ** postfields,
			 struct MBEventInfo event)
{
  if (event.time != NULL)
    add_chunk_to_postfields4(curl, postfields, "edit-event.time", event.time);

  return(0);
}

int add_urls_to_event(CURL * curl, char ** postfields,
		      struct MBEventInfo event)
{
  char * lt_field_name;
  char * text_field_name;

  int i;

  char * ltfnf = "edit-event.url.%d.link_type_id";
  char * tfnf = "edit-event.url.%d.text";

  lt_field_name = malloc(strlen(ltfnf) + event.url_count);
  text_field_name = malloc(strlen(tfnf) + event.url_count);

  for (i = 0; i < event.url_count && event.urls; i++)
  {
    sprintf(lt_field_name, ltfnf, i);
    sprintf(text_field_name, tfnf, i);

    add_chunk_to_postfields5(curl, postfields,
			     lt_field_name, event.urls[i].type);

    add_chunk_to_postfields4(curl, postfields,
			     text_field_name, event.urls[i].address);
  }

  free(lt_field_name);
  free(text_field_name);
  return(0);
}

/*
adds relationships from the "relationships" field, as well as the venue, if
given

assumes that postfields is a pointer to malloc'd heap array, and may
do a free(postfields) or a realloc
*/
int add_relationships_to_event(CURL * curl, char ** postfields,
				  struct MBEventInfo event)
{
  int i = 0;
  char * bwfield;
  char * typefield;
  char * targetfield;

  char * bwfieldformat = "edit-event.rel.%d.backward";
  char * typefieldformat = "edit-event.rel.%d.link_type_id";
  char * targetfieldformat = "edit-event.rel.%d.target";
  bwfield = malloc(strlen(bwfieldformat) + event.relationship_count);
  typefield = malloc(strlen(typefieldformat) + event.relationship_count);
  targetfield = malloc(strlen(targetfieldformat) + event.relationship_count);

  for ( ; i < event.relationship_count && event.relationships; i++)
  {
    if (event.relationships[i].backward)
    {
      add_numbered_chunk_to_postfields2(curl, postfields, bwfield,
					bwfieldformat, i, "1");
    }

    sprintf(typefield, typefieldformat, i);
    add_chunk_to_postfields5(curl, postfields,
			     typefield, event.relationships[i].type);

    sprintf(targetfield, targetfieldformat, i);
    add_chunk_to_postfields4(curl, postfields, targetfield,
			     event.relationships[i].musicbrainzID);
  }

  if (strlen(event.venue) == MBID_length)
  {
    sprintf(typefield, typefieldformat, i);
    add_chunk_to_postfields4(curl, postfields, typefield, "794");
    sprintf(targetfield, targetfieldformat, i);
    add_chunk_to_postfields4(curl, postfields, targetfield, event.venue);
  }
  
  return(0);
}

int lookup_area(struct MBState * mbs, MBID area)
{
  char * area_url;
  char * resultpage = NULL;
  long int status_code = 0;
  CURLcode ret;
  struct curl_slist * headers = NULL;
  json_t * data;
  json_t * id_thing;
  json_t * et_thing;
  json_error_t error;
  int id_int;

  char * area_url_format = "http://musicbrainz.org/ws/js/entity/%s";

  area_url = malloc(strlen(area_url_format) + MBID_length);

  sprintf(area_url, area_url_format, area);

  headers = curl_slist_append(headers, "Accept: application/json");
  if (headers == NULL)
  {
    printf("something went wrong with curl_slist_append\n");
  }

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_HTTPGET, 1);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_URL, area_url);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_HTTPHEADER, headers);

  ret = curl_easy_perform(mbs->curl_handle);

  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_RESPONSE_CODE,
			  &status_code);

  curl_slist_free_all(headers);
  curl_easy_reset(mbs->curl_handle);

  data = json_loads(resultpage, 0, &error);
  id_thing = json_object_get(data, "id");
  et_thing = json_object_get(data, "entityType");

  if(!json_is_string(et_thing))
  {
    printf("entityType entry is not a string\n");
    json_decref(data);
    return(-1);
  }

  if(strcmp("area", json_string_value(et_thing)) != 0)
  {
    printf("not an area\n");
    json_decref(data);
    return(-1);
  }

  switch(json_typeof(id_thing)) {
    case JSON_STRING:
      id_int = atoi(json_string_value(id_thing));
      json_decref(data);
      break;
    case JSON_INTEGER:
      id_int = json_integer_value(id_thing);
      json_decref(data);
      break;
    case JSON_OBJECT:
      json_decref(data);
      return(-1);
    default :
      json_decref(data);
      return(-1);
  }

  return(id_int);
}

void mb_init_ArtistInfo(struct MBArtistInfo * ai)
{
  /*
  ai->artist_name = ai->sort_name = ai->disambiguation = ai->urls = NULL;
  */
  ai->artist_name = NULL;
  ai->sort_name = NULL;
  ai->disambiguation = NULL;
  ai->urls = NULL;
  ai->area_id = 0;
  ai->type = Unknown;
  ai->gender = genderUnknown;
  ai->relationships = NULL;
  ai->not_dupe = ai->url_count = ai->relationship_count = 0;
}

void mb_init_Place(struct MBPlace * p)
{
  /*
  p->artist_name = p->sort_name = p->disambiguation = p->urls = NULL;
  */
  p->name = NULL;
  p->disambiguation = NULL;
  p->type = placeUnknown;
  p->coordinates = NULL;
  p->address = NULL;
  p->area_id = 0;
  p->urls = NULL;
  p->relationships = NULL;
  p->not_dupe = p->url_count = p->relationship_count = 0;
}

void mb_init_EventInfo(struct MBEventInfo * ei)
{
  ei->name = NULL;
  ei->disambiguation = NULL;
  ei->setlist = NULL;
  ei->type = 0;
  ei->begin_year = 0;
  ei->begin_month = 0;
  ei->begin_day = 0;
  ei->end_year = 0;
  ei->end_month = 0;
  ei->end_day = 0;
  ei->time = NULL;
  ei->venue[0] = '\0';
  ei->url_count = 0;
  ei->urls = NULL;
  ei->relationship_count = 0;
  ei->relationships = NULL;
}

int get_mbid_from_artist_url(char * url, MBID * mbidp)
{
  int slash;
  char * url2;
  char * url3;
  int i;

  if(url == NULL)
  {
    (*mbidp)[0] = '\0';
    return(0);
  }

  url2 = malloc(strlen(url) + 1);
  strcpy(url2, url);
  
  for(i = strlen(url2) - 1 ; url2[i] == '/'; i--)
    url2[i] = 0;

  url3 = strrchr(url2, '/') + 1;

  strncpy(*mbidp, url3, MBID_length);
  (*mbidp)[MBID_length] = 0;
  free(url2);
}

/*
name and sort_name fields of "artist" argument are required
*/
int mb_do_artist(CURL * ch, char * url, struct MBArtistInfo artist)
{
  char * postfields;

  CURLcode ret;
  long int status_code = 0;
  char * redirect_url;
  char * resultpage = NULL;
  int i;

  postfields = build_artist_name_post_string(ch, artist);

  add_disambig_to_artist(ch, &postfields, artist);
  add_type_to_artist(ch, &postfields, artist);
  add_gender_to_artist(ch, &postfields, artist);
  add_area_to_artist(ch, &postfields, artist);
  add_urls_to_artist(ch, &postfields, artist);
  add_relationships_to_artist(ch, &postfields, artist);

  ret = curl_easy_setopt(ch, CURLOPT_URL, url);

  /*
  ret = curl_easy_setopt(ch, CURLOPT_HEADERDATA, &header_info);
  ret = curl_easy_setopt(ch, CURLOPT_HEADERFUNCTION,
			 &add_artist_header_callback);
  */
  ret = curl_easy_setopt(ch, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(ch, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);
  ret = curl_easy_setopt(ch, CURLOPT_POSTFIELDS, postfields);

  ret = curl_easy_perform(ch);
  if(ret != CURLE_OK)
  {
    printf("curl_easy_perform returned %d\n", ret);
  }

  ret = curl_easy_getinfo(ch, CURLINFO_RESPONSE_CODE,
			  &status_code);
  ret = curl_easy_getinfo(ch, CURLINFO_REDIRECT_URL,
			  &redirect_url);

  printf("  status_code is %ld\n", status_code);
  printf("  redirect_url is %s\n", redirect_url);

  if(status_code == 200 || status_code == 500)
  {
    printf("page:\n");
    puts(resultpage);
  }

  CURLoption opts[] = {CURLOPT_WRITEDATA, CURLOPT_WRITEFUNCTION,
		       CURLOPT_POSTFIELDS};
  nullify_curl_options(ch, 3, opts);
  free(resultpage);
  free((void *) postfields);
}

int mb_add_artist(struct MBState * mbs, struct MBArtistInfo artist,
		  MBID* artist_mbidp)
{
  char * postfields;
  CURLcode ret;
  long int status_code = 0;
  char * redirect_url;
  char * resultpage = NULL;
  char * header_info = NULL;
  int i;

  postfields = build_artist_name_post_string(mbs->curl_handle, artist);

  add_type_to_artist(mbs->curl_handle, &postfields, artist);
  add_disambig_to_artist(mbs->curl_handle, &postfields, artist);
  add_urls_to_artist(mbs->curl_handle, &postfields, artist);
  add_relationships_to_artist(mbs->curl_handle, &postfields, artist);
  add_area_to_artist(mbs->curl_handle, &postfields, artist);

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_URL,
			 "https://musicbrainz.org/artist/create");

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_POSTFIELDS, postfields);

  ret = curl_easy_perform(mbs->curl_handle);
  if(ret != CURLE_OK)
  {
    printf("curl_easy_perform returned %d\n", ret);
  }

  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_RESPONSE_CODE,
			  &status_code);
  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_REDIRECT_URL,
			  &redirect_url);

  if(status_code == 200 || (status_code > 499 & status_code < 599))
  {
    FILE * pageout;

    free(resultpage);
    free((void *) postfields);
    return(1);
  }

  get_mbid_from_artist_url(redirect_url, artist_mbidp);

  free(resultpage);
  free((void *) postfields);
  return(0);
}

int mb_update_artist(struct MBState * mbs, char musicbrainzID[],
		     struct MBArtistInfo artist)
{
  int ret_int;
  char * url ;
  char * resultpage = NULL;

  char * urlformat = "https://musicbrainz.org/artist/%s/edit";
  url = malloc(strlen(urlformat) + strlen(musicbrainzID));
  sprintf(url, urlformat, musicbrainzID);

  ret_int = mb_do_artist(mbs->curl_handle, url, artist);
  free(url);
}

int mb_add_place(struct MBState * mbs, struct MBPlace place,
		    MBID* place_mbidp)
{
  char * postfields;

  CURLcode ret;
  long int status_code = 0;
  char * redirect_url;
  char * resultpage = NULL;

  char * subs;

  postfields = build_place_name_post_string(mbs->curl_handle, place);
  add_disambiguation_to_place(mbs->curl_handle, &postfields, place);
  /*
  add_dupe_to_place(mbs->curl_handle, &postfields, place);
  */
  add_type_to_place(mbs->curl_handle, &postfields, place);
  add_coordinates_to_place(mbs->curl_handle, &postfields, place);
  add_address_to_place(mbs->curl_handle, &postfields, place);
  add_area_id_to_place(mbs->curl_handle, &postfields, place);
  add_urls_to_place(mbs->curl_handle, &postfields, place);
  add_relationships_to_place(mbs->curl_handle, &postfields, place);

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_URL,
			 "https://musicbrainz.org/place/create");

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_POSTFIELDS, postfields);

  ret = curl_easy_perform(mbs->curl_handle);

  if(ret != CURLE_OK)
  {
    printf("L604: curl_easy_perform returned %d\n", ret);
  }

  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_RESPONSE_CODE,
			  &status_code);
  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_REDIRECT_URL,
			  &redirect_url);

  for(subs = redirect_url; subs[0] != '\0'; subs = subs + 1) {
    if (strncmp(subs, "/place/", 7) == 0)
    {
      subs = subs + 7;
      break;
    }
  }

  strncpy(*place_mbidp, subs, MBID_length);
  (*place_mbidp)[MBID_length] = 0;
  curl_easy_reset(mbs->curl_handle);
  return(0);
}

/*
currently assumes that event has one and only one venue
*/
int mb_add_event(struct MBState * mbs, struct MBEventInfo event,
		 MBID* event_mbidp)
{
  char * postfields;

  CURLcode ret;
  long int status_code = 0;
  char * redirect_url;
  char * resultpage = NULL;

  postfields = build_event_name_post_string(mbs->curl_handle, event);
  add_disambig_to_event(mbs->curl_handle, &postfields, event);
  add_setlist_to_event(mbs->curl_handle, &postfields, event);

  add_type_to_event(mbs->curl_handle, &postfields, event);
  add_begin_year_to_event(mbs->curl_handle, &postfields, event);
  add_begin_month_to_event(mbs->curl_handle, &postfields, event);
  add_begin_day_to_event(mbs->curl_handle, &postfields, event);
  add_end_year_to_event(mbs->curl_handle, &postfields, event);
  add_end_month_to_event(mbs->curl_handle, &postfields, event);
  add_end_day_to_event(mbs->curl_handle, &postfields, event);
  add_time_to_event(mbs->curl_handle, &postfields, event);
  add_urls_to_event(mbs->curl_handle, &postfields, event);
  add_relationships_to_event(mbs->curl_handle, &postfields, event);

  printf("L981: postfields is |%s|\n", postfields);

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_URL,
			 "https://musicbrainz.org/event/create");

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEDATA, &resultpage);
  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_WRITEFUNCTION,
			 &build_string_write_callback);

  ret = curl_easy_setopt(mbs->curl_handle, CURLOPT_POSTFIELDS, postfields);

  ret = curl_easy_perform(mbs->curl_handle);

  if(ret != CURLE_OK)
  {
    printf("L604: curl_easy_perform returned %d\n", ret);
  }

  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_RESPONSE_CODE,
			  &status_code);
  ret = curl_easy_getinfo(mbs->curl_handle, CURLINFO_REDIRECT_URL,
			  &redirect_url);

  printf("status_code is %ld\n", status_code);
  printf("redirect_url is %s\n", redirect_url);

  if(status_code == 200 || status_code == 500)
  {
    FILE * pageout;

    pageout = fopen("add_event.output.html", "w");
    fputs(resultpage, pageout);
    fclose(pageout);

    printf("page result is in add_event.output.html\n");
    
    /*
    puts(resultpage);
    */
  }

  get_mbid_from_artist_url(redirect_url, event_mbidp);
  printf("1027 of musicbrainz-lib.c: mbid is %s\n", *event_mbidp);

  free(resultpage);
  free((void *) postfields);

}
