#include <curl/curl.h>

struct MBState {
  CURL * curl_handle;
  int logged_in;
};

struct MBURL {
  int type;
  char * address;
};

#define MBID_length 36
typedef char MBID[MBID_length + 1] ;

struct MBRelationship {
  int type;
  int backward;    /* boolean */
  MBID musicbrainzID;
};

enum artist_type {
  Unknown = 0,
  Person = 1,
  Group = 2,
  Other = 3,
  Character = 4,
  Orchestra = 5,
  Choir = 6
};

enum place_type {
  placeUnknown = 0,
  placeStudio = 1,
  placeVenue = 2,
  placeOther = 3,
  placeStadium = 4,
  placeIndoorArena = 5,
  placeReligiousBuilding = 6
};

enum event_type {
  eventUnknown = 0,
  eventConcert = 1,
  eventFestival = 2,
  eventLaunch = 3,
  eventConvention = 4,
  eventMasterclass = 5
};

enum genderEnum {
  genderUnknown = 0,
  genderMale = 1,
  genderFemale = 2,
  genderOther = 3
};

struct MBArtistInfo {
  char * artist_name ;   /* required */
  char * sort_name ;      /* required */
  char * disambiguation ;
  int not_dupe ;         /* 1 means "I confirm this is not duplicate entry" */
  enum artist_type type; /* if you exclude this, Musicbrainz will set to null*/
  enum genderEnum gender;/* if you exclude this, Musicbrainz will set to null*/
  int area_id ;
  int url_count ;
  struct MBURL * urls;
  int relationship_count;
  struct MBRelationship * relationships;
};

struct MBPlace {
  char * name ;   /* required */
  char * disambiguation ;
  int not_dupe ;         /* 1 means "I confirm this is not duplicate entry" */
  enum place_type type; /* if you exclude this, Musicbrainz will set to null?*/
  char * coordinates ;
  char * address ;
  int area_id ;
  int url_count ;
  struct MBURL * urls;
  int relationship_count;
  struct MBRelationship * relationships;
};

struct MBEventInfo {
  char * name;
  char * disambiguation;
  char * setlist;
  enum event_type type;
  int begin_year;
  int begin_month;
  int begin_day;
  int end_year;
  int end_month;
  int end_day;
  char * time;
  MBID venue;
  int url_count ;
  struct MBURL * urls;
  int relationship_count;
  struct MBRelationship * relationships;
};

void mb_global_initialize();

struct MBState * mb_initialize(void);

struct MBState * mbstate_alloc(void);

void mbstate_initialize(struct MBState * mbs);

void mb_init_ArtistInfo(struct MBArtistInfo * ai);

void mb_init_Place(struct MBPlace * p);

int lookup_area(struct MBState * mbs, MBID area);

void mb_init_EventInfo(struct MBEventInfo * ei);

int mb_login(struct MBState * mbs, char* username, char* password);

int mb_add_artist(struct MBState * mbs, struct MBArtistInfo artist,
		  MBID* artist_mbid);

int mb_update_artist(struct MBState * mbs, char musicbrainzID[],
		     struct MBArtistInfo artist);

/*
  returns a pointer to a dynamically allocated string
  the pointer should be freed by the calling function
*/
int mb_add_place(struct MBState * mbs, struct MBPlace place,
		    MBID* place_mbidp);

int mb_add_event(struct MBState * mbs, struct MBEventInfo event,
		 MBID* event_mbidp);
